﻿using System;
using System.Collections.Generic;
using System.Diagnostics;
using System.Linq;
using System.Text;
using System.Threading;
#if SILVERLIGHT
using System.Windows;
#endif

namespace Lobs.ApplicationModel
{
  /// <summary>
  /// Root application object.
  /// </summary>
  public class ApplicationProcess
  {
    /// <summary>
    /// Initializes a new instance of the <see cref="ApplicationProcess"/> class with the specified data model implementation.
    /// </summary>
    ApplicationProcess(DataModel.IDataProvider dataModel)
    {
      DataProvider = dataModel;
    }

    //public static ApplicationProcess Create(DataModel.IDataProvider dataModel)
    //{
    //  var appProcess = new ApplicationProcess(dataModel);
    //  return appProcess;
    //}

    /// <summary>
    /// Initializes the application process singleton.
    /// </summary>
    public static void Initialize(string providerName)
    {
      if (_appProcess != null)
        throw new InvalidOperationException("Only one ApplicationProcess must exist.");

      DataModel.IDataProvider dataProvider = DataModel.DataProviderFactory.GetProvider(providerName);

      _appProcess = new ApplicationProcess(dataProvider);
    }

    /// <summary>
    /// Deinitializes the application process singleton. Used for unit tests only.
    /// </summary>
    public static void DeInitialize()
    {
      if (_appProcess != null)
        _appProcess = null;
    }

    /// <summary>
    /// Gets the current application model singleton.
    /// </summary>
    public static ApplicationProcess Current
    {
      get { return _appProcess; }
    }
    static ApplicationProcess _appProcess;

    public static void AcquireAppModel()
    {
      EnsureBackgroundThread();
      Monitor.Enter(_appProcess);
    }

    public static void ReleaseAppModel()
    {
      Monitor.Exit(_appProcess);
    }

    public DataModel.IDataProvider DataProvider { get; set; }

    /// <summary>
    /// Ensures the calling thread is not the UI thread.
    /// </summary>
    [Conditional("SILVERLIGHT")]
    internal static void EnsureBackgroundThread()
    {
#if SILVERLIGHT
      if (Deployment.Current.CheckAccess())
        throw new InvalidOperationException("Must not be called from UI thread.");
#endif
    }

    /// <summary>
    /// Ensures the calling thread is not the UI thread.
    /// </summary>
    [Conditional("SILVERLIGHT")]
    internal static void EnsureForegroundThread()
    {
#if SILVERLIGHT
      if (!Deployment.Current.CheckAccess())
        throw new InvalidOperationException("Must be called from UI thread.");
#endif
    }


    public void InitializeKeyTables()
    {
      LoadKeyTables_V1();
    }

    void LoadKeyTables_V1()
    {
      if (!_keyTablesInitialized)
      {
        _keyTablesInitialized = true;
#if SILVERLIGHT
#else
        _keyTables.SetCategories(DataProvider.LoadCategories());
        _keyTables.SetShippers(DataProvider.LoadShippers());
        _keyTablesInitialized = true;
#endif
      }
    }

    public bool KeyTablesInitialized
    {
      get { return _keyTablesInitialized; }
    }

    private bool _keyTablesInitialized;

    /// <summary>
    /// Gets key tables like Categories.
    /// </summary>
    public KeyTables KeyTables
    {
      get { return _keyTables; }
    }

    readonly KeyTables _keyTables = new KeyTables();





    // TODO: nur speichern, was sich geändert hat! => weitere Methoden?
    /// <summary>
    /// Saves all data to the default data provider.
    /// </summary>
    public void SaveData()
    {
      //  SaveData(DataProvider);
    }

    /// <summary>
    /// Ensures that the data has been loaded from the data provider.
    /// </summary>
    public void EnsureData()
    {
      //if (Krankenkassen.Count == 0)
      //  LoadData();
    }

    /// <summary>
    /// Determines whether data was already loaded from the data provider.
    /// </summary>
    public bool HasData()
    {
      return false; // Krankenkassen.Count > 0;
    }

    /// <summary>
    /// Sample: Application provides a list of employess.
    /// </summary>
    public List<Employee> Employees
    {
      get
      {
        if (_employees == null)
        {
          _employees = new List<Employee>();

          // Load from data provider
          var data = DataProvider.LoadEmployees();

          // Create application model objects
          foreach (var dmEmployee in data)
            _employees.Add(Employee.FromDataModel(dmEmployee));
        }
        return _employees;
      }
    }

    List<Employee> _employees;

    /// <summary>
    /// Loads all employees.
    /// </summary>
    public List<Employee> LoadEmployees()
    {
      // Load from data provider
      var data = DataProvider.LoadEmployees();

      // Create application model objects
      var list = new List<Employee>();
      foreach (var dmEmployee in data)
        list.Add(Employee.FromDataModel(dmEmployee));

      return list;
    }

    ///// <summary>
    ///// Load photo of employee.
    ///// </summary>
    //public List<Employee> LoadEmployees()
    //{
    //  // Load from data provider
    //  var data = DataProvider.LoadEmployees();

    //  // Create application model objects
    //  var list = new List<Employee>();
    //  foreach (var dmEmployee in data)
    //    list.Add(Employee.FromDataModel(dmEmployee));

    //  return list;
    //}

    /// <summary>
    /// Loads all products.
    /// </summary>
    public List<Product> LoadProducts()
    {
      // Load from data provider
      var data = DataProvider.LoadProducts();

      // Create application model objects
      var list = new List<Product>();
      foreach (var dmProduct in data)
        list.Add(Product.FromDataModel(dmProduct));

      return list;
    }


#if delete
    /// <summary>
    /// Frees the data (forcing a reload).
    /// </summary>
    public void ClearData()
    {
      if (Krankenkassen.Count > 0)
      {
        EmptyLists();
      }
    }

    public List<Krankenkasse> LoadKrankenkassenÜbersicht(Func<Krankenkasse, bool> filter)
    {
      EnsureData();
#if true
      return (from krankenkasse in Krankenkassen where filter(krankenkasse) select krankenkasse).ToList();
#else
      var result = new List<Krankenkasse>();
      foreach (var kassen in Krankenkassen)
      {
        if (filter(kassen))
          result.Add(kassen);
      }
      return result;
#endif
    }

    public List<Tarif> LoadTarifeÜbersicht(Func<Tarif, bool> filter)
    {
      EnsureData();
      return (from tarif in Tarife where filter(tarif) select tarif).ToList();
    }

    public List<Dokument> LoadDokumenteÜbersicht(Func<Dokument, bool> filter)
    {
      EnsureData();
      return (from dokument in Dokumente where filter(dokument) select dokument).ToList();
    }

    /// <summary>
    /// Gets the specified Krankenkasse.
    /// </summary>
    public Krankenkasse GetKrankenkasse(string id)
    {
      return (from krankenkasse in Krankenkassen where krankenkasse.ID == id select krankenkasse).FirstOrDefault();
    }

    /// <summary>
    /// Gets the specified Tarif.
    /// </summary>
    public Tarif GetTarif(string tarifId)
    {
      // Warum IsNullOrEmpty?
      // Siehe GetKrankenkasse
      return (from tarif in Tarife where !String.IsNullOrEmpty(tarif.ID) && tarif.ID == tarifId select tarif).FirstOrDefault();
    }

    /// <summary>
    /// Gets the specified Bevölkerungsgruppe.
    /// </summary>
    public Bevoelkerungsgruppe GetBevoelkerungsgruppe(string bevoelkerungsgruppeId)
    {
      foreach (Bevoelkerungsgruppe bevoelkerungsgruppe in Bevoelkerungsgruppen)
      {
        if (!String.IsNullOrEmpty(bevoelkerungsgruppe.ID) && bevoelkerungsgruppe.ID == bevoelkerungsgruppeId)
          return bevoelkerungsgruppe;
      }
      return null;
    }

    /// <summary>
    /// Gets a specific Regionalebene
    /// </summary>
    public Regionalebene GetRegionalebene(string regionalebeneId)
    {
      foreach (Regionalebene regionalebene in Regionalebenen)
      {
        if (!String.IsNullOrEmpty(regionalebene.ID) && regionalebene.ID == regionalebeneId)
          return regionalebene;
      }
      return null;
    }

    /// <summary>
    /// Gets a specific Tarifart
    /// </summary>
    public Tarifart GetTarifart(string tarifartId)
    {
      foreach (Tarifart tarifart in Tarifarten)
      {
        if (!String.IsNullOrEmpty(tarifart.ID) && tarifart.ID == tarifartId)
          return tarifart;
      }
      return null;
    }

    /// <summary>
    /// Gets a specific Kassenart
    /// </summary>
    public Kassenart GetKassenart(string kassenartId)
    {
      foreach (Kassenart kassenart in Kassenarten)
      {
        if (!String.IsNullOrEmpty(kassenart.ID) && kassenart.ID == kassenartId)
          return kassenart;
      }
      return null;
    }

    /// <summary>
    /// Gets a list of Versicherungsart for a given Tarifart
    /// </summary>
    public List<Versicherungsart> GetVersicherungsarten(Tarifart tarifart)
    {
      var versicherungsarten = new List<Versicherungsart>();

      if (tarifart != null && !String.IsNullOrEmpty(tarifart.ID))
      {
        foreach (Versicherungsart versicherungsart in Versicherungsarten)
        {
          foreach (Tarif tarif in Tarife)
          {
            if (tarif.Versicherungsart != null && tarif.Versicherungsart.Equals(versicherungsart) && 
                tarif.Tarifart != null && tarif.Tarifart.Equals(tarifart))
            {
              versicherungsarten.Add(versicherungsart);
              break;
            }
          }
        }
      }
      return versicherungsarten;
    }

    /// <summary>
    /// Gets a specific Versicherungsart
    /// </summary>
    public Versicherungsart GetVersicherungsart(string versicherungsartId)
    {
      foreach (Versicherungsart versicherungsart in Versicherungsarten)
      {
        if (!String.IsNullOrEmpty(versicherungsart.ID) && versicherungsart.ID == versicherungsartId)
          return versicherungsart;
      }
      return null;
    }

    /// <summary>
    /// Gets a specific Dokument
    /// </summary>
    public Dokument GetDokument(string dokumentId)
    {
      foreach (Dokument dokument in Dokumente)
      {
        if (!String.IsNullOrEmpty(dokument.ID) && dokument.ID == dokumentId)
          return dokument;
      }
      return null;
    }

    /// <summary>
    /// Gets a specific Dokument
    /// </summary>
    public Dokument GetDokumentByFilename(string filename)
    {
      foreach (Dokument dokument in Dokumente)
      {
        if (!String.IsNullOrEmpty(dokument.Filename) && String.Compare(dokument.Filename, filename) == 0)
          return dokument;
      }
      return null;
    }


    public bool Serialize()
    {
#if false
      string fileName = "Serialized.xml";
      System.IO.IsolatedStorage.IsolatedStorageFile isoStore = System.IO.IsolatedStorage.IsolatedStorageFile.GetUserStoreForApplication();
      System.IO.IsolatedStorage.IsolatedStorageFileStream stream = new System.IO.IsolatedStorage.IsolatedStorageFileStream(fileName, System.IO.FileMode.Create, isoStore);
#endif

#if SILVERLIGHT
      return false;
#else
      string fileName = "..\\..\\..\\..\\XML\\AMSerialized.xml";
      System.IO.FileStream stream = new System.IO.FileStream(fileName, System.IO.FileMode.Create);


      System.Xml.Serialization.XmlSerializer serializer = new System.Xml.Serialization.XmlSerializer(this.GetType());
      serializer.Serialize(stream, this);

      stream.Flush();
      stream.Close();

      return true;
#endif
    }

    void EmptyLists()
    {
      Krankenkassen.Clear();
      KrankenkasseRegionalebenen.Clear();
      Kassenarten.Clear();
      Versicherungsarten.Clear();
      Dokumente.Clear();
      Tarife.Clear();
      Tarifarten.Clear();
      Regionalebenen.Clear();
      Bevoelkerungsgruppen.Clear();
      RegionalebeneBevoelkerungen.Clear();
    }

    public List<Krankenkasse> Krankenkassen = new List<Krankenkasse>();
    public List<KrankenkasseRegionalebene> KrankenkasseRegionalebenen = new List<KrankenkasseRegionalebene>();
    public List<Kassenart> Kassenarten = new List<Kassenart>();
    public List<Versicherungsart> Versicherungsarten = new List<Versicherungsart>();
    public List<Dokument> Dokumente = new List<Dokument>();
    public List<Tarif> Tarife = new List<Tarif>();
    public List<Tarifart> Tarifarten = new List<Tarifart>();
    public List<Regionalebene> Regionalebenen = new List<Regionalebene>();
    public List<Bevoelkerungsgruppe> Bevoelkerungsgruppen = new List<Bevoelkerungsgruppe>();
    public List<RegionalebeneBevölkerung> RegionalebeneBevoelkerungen = new List<RegionalebeneBevölkerung>();
#endif

  }
}
