﻿using System;
using System.Collections.Generic;
using System.ComponentModel;
using System.Threading;
#if SILVERLIGHT
using System.Windows;
#endif

namespace Frmwrk.ComponentModel
{
  /// <summary>
  /// Base class for both service functions and service actions.
  /// </summary>
  public class AsyncService   // rename to: SynchronizedAsyncCall
  {
    public static AsyncServiceFunction<TResult> CreateFunction<TResult>(Func<TResult> function)
    {
      return new AsyncServiceFunction<TResult>(function);
    }

    public static AsyncServiceFunction<T1, TResult> CreateFunction<T1, TResult>(Func<T1, TResult> function)
    {
      return new AsyncServiceFunction<T1, TResult>(function);
    }

    public static AsyncServiceFunction<T1, T2, TResult> CreateFunction<T1, T2, TResult>(Func<T1, T2, TResult> function)
    {
      return new AsyncServiceFunction<T1, T2, TResult>(function);
    }

    public static AsyncServiceFunction<T1, T2, T3, TResult> CreateFunction<T1, T2, T3, TResult>(Func<T1, T2, T3, TResult> function)
    {
      return new AsyncServiceFunction<T1, T2, T3, TResult>(function);
    }

    public static AsyncServiceFunction<T1, T2, T3, T4, TResult> CreateFunction<T1, T2, T3, T4, TResult>(Func<T1, T2, T3, T4, TResult> function)
    {
      return new AsyncServiceFunction<T1, T2, T3, T4, TResult>(function);
    }

    public static AsyncServiceFunction<T1, T2, T3, T4, T5, TResult> CreateFunction<T1, T2, T3, T4, T5, TResult>(Func<T1, T2, T3, T4, T5, TResult> function)
    {
      return new AsyncServiceFunction<T1, T2, T3, T4, T5, TResult>(function);
    }

    public static AsyncServiceAction CreateAction(Action action)
    {
      return new AsyncServiceAction(action);
    }

    public static AsyncServiceAction<T1> CreateAction<T1>(Action<T1> action)
    {
      return new AsyncServiceAction<T1>(action);
    }

    public static AsyncServiceAction<T1, T2> Create<T1, T2>(Action<T1, T2> action)
    {
      return new AsyncServiceAction<T1, T2>(action);
    }

    public static AsyncServiceAction<T1, T2, T3> CreateAction<T1, T2, T3>(Action<T1, T2, T3> action)
    {
      return new AsyncServiceAction<T1, T2, T3>(action);
    }

    public static AsyncServiceAction<T1, T2, T3, T4> CreateAction<T1, T2, T3, T4>(Action<T1, T2, T3, T4> action)
    {
      return new AsyncServiceAction<T1, T2, T3, T4>(action);
    }

    //public static AsyncServiceAction<T1, T2, T3, T4, T5 > CreateAction<T1, T2, T3, T4, T5>(Action<T1, T2, T3, T4, T5> action)
    //{
    //  return new AsyncServiceAction<T1, T2, T3, T4, T5>(action);
    //}

    /// <summary>
    /// Initializes a new instance of the <see cref="AsyncService"/> class.
    /// </summary>
    protected AsyncService()
    {
#if SILVERLIGHT
      if (!Deployment.Current.CheckAccess())
        throw new InvalidOperationException("Must only be used in the UI thread.");
#endif
      // Ensure that a SynchronizationContext exists for the UI thread
      _synchronizationContext = SynchronizationContext.Current;
      if (_synchronizationContext == null)
      {
        _synchronizationContext = new SynchronizationContext();
        SynchronizationContext.SetSynchronizationContext(_synchronizationContext);
      }
    }

    /// <summary>
    /// Gets the synchronization context. Used for sending ServiceCompleted event.
    /// </summary>
    protected SynchronizationContext SynchronizationContext
    {
      get { return _synchronizationContext; }
    }

    // SynchronizationContext of the UI thread
    readonly SynchronizationContext _synchronizationContext;
  }

  /// <summary>
  /// Base class of all service function classes.
  /// </summary>
  public class AsyncServiceFunctionBase<TResult> : AsyncService
  {
    /// <summary>
    /// Runs the service function within a new thread.
    /// </summary>
    protected void RunAsync(SendOrPostCallback callback)
    {
      if (ServiceCompleted == null)
        throw new InvalidOperationException("Must set ServiceCompleted before service function can be invoked.");

      ThreadPool.QueueUserWorkItem(x => callback(null));
    }

    protected void RaiseServiceCompleted(object sender, AsyncServiceCompletedEventArgs<TResult> e)
    {
      if (ServiceCompleted != null)
        ServiceCompleted(sender, e);
    }

    public event EventHandler<AsyncServiceCompletedEventArgs<TResult>> ServiceCompleted;
  }

  /// <summary>
  /// Base class of all service action classes.
  /// </summary>
  public class AsyncServiceActionBase : AsyncService
  {
    /// <summary>
    /// Runs the service action within a new thread.
    /// </summary>
    protected void RunAsync(SendOrPostCallback callback)
    {
      if (ServiceCompleted == null)
        throw new InvalidOperationException("Must set ServiceCompleted before service function can be invoked.");

      // TODO: use thread pool
      new Thread(() => callback(null)).Start();
    }

    protected void RaiseServiceCompleted(object sender, AsyncCompletedEventArgs e)
    {
      if (ServiceCompleted != null)
        ServiceCompleted(sender, e);
    }

    public event EventHandler<AsyncCompletedEventArgs> ServiceCompleted;
  }

  /// <summary>
  /// Encapsulates a service function that has no parameters and returns a value of the type specified by the TResult parameter.
  /// </summary>
  public class AsyncServiceFunction<TResult> : AsyncServiceFunctionBase<TResult>
  {
    public AsyncServiceFunction(Func<TResult> function)
    {
      _function = function;
    }

    public void InvokeServiceAsync()
    {
      InvokeServiceAsync(null);
    }

    public void InvokeServiceAsync(object userState)
    {
      RunAsync(obj => DoWork(userState));
    }

    void DoWork(object userState)
    {
      TResult result = default(TResult);
      Exception error = null;
      try
      {
        result = _function();
      }
      catch (Exception ex)
      {
        error = ex;
      }
      SynchronizationContext.Send(obj => RaiseServiceCompleted(null, new AsyncServiceCompletedEventArgs<TResult>(result, error, false, userState)), null);
    }

    readonly Func<TResult> _function;
  }

  /// <summary>
  /// Encapsulates a service function that has one  parameters and returns a value of the type specified by the TResult parameter.
  /// </summary>
  public class AsyncServiceFunction<T, TResult> : AsyncServiceFunctionBase<TResult>
  {
    public AsyncServiceFunction(Func<T, TResult> function)
    {
      _function = function;
    }

    public void InvokeServiceAsync(T arg)
    {
      InvokeServiceAsync(arg, null);
    }

    public void InvokeServiceAsync(T arg, object userState)
    {
      RunAsync(obj => DoWork(arg, userState));
    }

    void DoWork(T arg, object userState)
    {
      TResult result = default(TResult);
      Exception error = null;
      try
      {
        result = _function(arg);
      }
      catch (Exception ex)
      {
        error = ex;
      }
      SynchronizationContext.Send(obj => RaiseServiceCompleted(null, new AsyncServiceCompletedEventArgs<TResult>(result, error, false, userState)), null);
    }

    readonly Func<T, TResult> _function;
  }

  /// <summary>
  /// Encapsulates a service function that has two parameters and returns a value of the type specified by the TResult parameter.
  /// </summary>
  public class AsyncServiceFunction<T1, T2, TResult> : AsyncServiceFunctionBase<TResult>
  {
    public AsyncServiceFunction(Func<T1, T2, TResult> function)
    {
      _function = function;
    }

    public void InvokeServiceAsync(T1 arg1, T2 arg2)
    {
      InvokeServiceAsync(arg1, arg2, null);
    }

    public void InvokeServiceAsync(T1 arg1, T2 arg2, object userState)
    {
      RunAsync(obj => DoWork(arg1, arg2, userState));
    }

    void DoWork(T1 arg1, T2 arg2, object userState)
    {
      TResult result = default(TResult);
      Exception error = null;
      try
      {
        result = _function(arg1, arg2);
      }
      catch (Exception ex)
      {
        error = ex;
      }
      SynchronizationContext.Send(obj => RaiseServiceCompleted(null, new AsyncServiceCompletedEventArgs<TResult>(result, error, false, userState)), null);
    }

    readonly Func<T1, T2, TResult> _function;
  }

  /// <summary>
  /// Encapsulates a service function that has three parameters and returns a value of the type specified by the TResult parameter.
  /// </summary>
  public class AsyncServiceFunction<T1, T2, T3, TResult> : AsyncServiceFunctionBase<TResult>
  {
    public AsyncServiceFunction(Func<T1, T2, T3, TResult> function)
    {
      _function = function;
    }

    public void InvokeServiceAsync(T1 arg1, T2 arg2, T3 arg3)
    {
      InvokeServiceAsync(arg1, arg2, arg3, null);
    }

    public void InvokeServiceAsync(T1 arg1, T2 arg2, T3 arg3, object userState)
    {
      RunAsync(obj => DoWork(arg1, arg2, arg3, userState));
    }

    void DoWork(T1 arg1, T2 arg2, T3 arg3, object userState)
    {
      TResult result = default(TResult);
      Exception error = null;
      try
      {
        result = _function(arg1, arg2, arg3);
      }
      catch (Exception ex)
      {
        error = ex;
      }
      SynchronizationContext.Send(obj => RaiseServiceCompleted(null, new AsyncServiceCompletedEventArgs<TResult>(result, error, false, userState)), null);
    }

    readonly Func<T1, T2, T3, TResult> _function;
  }

  /// <summary>
  /// Encapsulates a service function that has four parameters and returns a value of the type specified by the TResult parameter.
  /// </summary>
  public class AsyncServiceFunction<T1, T2, T3, T4, TResult> : AsyncServiceFunctionBase<TResult>
  {
    public AsyncServiceFunction(Func<T1, T2, T3, T4, TResult> function)
    {
      _function = function;
    }

    public void InvokeServiceAsync(T1 arg1, T2 arg2, T3 arg3, T4 arg4)
    {
      InvokeServiceAsync(arg1, arg2, arg3, arg4, null);
    }

    public void InvokeServiceAsync(T1 arg1, T2 arg2, T3 arg3, T4 arg4, object userState)
    {
      RunAsync(obj => DoWork(arg1, arg2, arg3, arg4, userState));
    }

    void DoWork(T1 arg1, T2 arg2, T3 arg3, T4 arg4, object userState)
    {
      TResult result = default(TResult);
      Exception error = null;
      try
      {
        result = _function(arg1, arg2, arg3, arg4);
      }
      catch (Exception ex)
      {
        error = ex;
      }
      SynchronizationContext.Send(obj => RaiseServiceCompleted(null, new AsyncServiceCompletedEventArgs<TResult>(result, error, false, userState)), null);
    }

    readonly Func<T1, T2, T3, T4, TResult> _function;
  }

  /// <summary>
  /// Encapsulates a method that has five parameters and returns a value of the type specified by the TResult parameter. Not yet defined in .NET framework.
  /// </summary>
  public delegate TResult Func<T1, T2, T3, T4, T5, TResult>(T1 arg1, T2 arg2, T3 arg3, T4 arg4, T5 arg5);

  /// <summary>
  /// Encapsulates a service function that has five parameters and returns a value of the type specified by the TResult parameter.
  /// </summary>
  public class AsyncServiceFunction<T1, T2, T3, T4, T5, TResult> : AsyncServiceFunctionBase<TResult>
  {
    public AsyncServiceFunction(Func<T1, T2, T3, T4, T5, TResult> function)
    {
      _function = function;
    }

    public void InvokeServiceAsync(T1 arg1, T2 arg2, T3 arg3, T4 arg4, T5 arg5)
    {
      InvokeServiceAsync(arg1, arg2, arg3, arg4, arg5, null);
    }

    public void InvokeServiceAsync(T1 arg1, T2 arg2, T3 arg3, T4 arg4, T5 arg5, object userState)
    {
      RunAsync(obj => DoWork(arg1, arg2, arg3, arg4, arg5, userState));
    }

    void DoWork(T1 arg1, T2 arg2, T3 arg3, T4 arg4, T5 arg5, object userState)
    {
      TResult result = default(TResult);
      Exception error = null;
      try
      {
        result = _function(arg1, arg2, arg3, arg4, arg5);
      }
      catch (Exception ex)
      {
        error = ex;
      }
      SynchronizationContext.Send(obj => RaiseServiceCompleted(null, new AsyncServiceCompletedEventArgs<TResult>(result, error, false, userState)), null);
    }

    readonly Func<T1, T2, T3, T4, T5, TResult> _function;
  }

  /// <summary>
  /// Encapsulates a service method that takes no parameters and does not return a value.
  /// </summary>
  public class AsyncServiceAction : AsyncServiceActionBase
  {
    public AsyncServiceAction(Action action)
    {
      _action = action;
    }

    public void InvokeServiceAsync()
    {
      InvokeServiceAsync(null);
    }

    public void InvokeServiceAsync(object userState)
    {
      RunAsync(obj => DoWork(userState));
    }

    void DoWork(object userState)
    {
      Exception error = null;
      try
      {
        _action();
      }
      catch (Exception ex)
      {
        error = ex;
      }
      SynchronizationContext.Send(obj => RaiseServiceCompleted(null, new AsyncCompletedEventArgs(error, false, userState)), null);
    }

    readonly Action _action;
  }

  /// <summary>
  /// Encapsulates a service method that takes one parameters and does not return a value.
  /// </summary>
  public class AsyncServiceAction<T> : AsyncServiceActionBase
  {
    public AsyncServiceAction(Action<T> action)
    {
      _action = action;
    }

    public void InvokeServiceAsync(T arg)
    {
      InvokeServiceAsync(arg, null);
    }

    public void InvokeServiceAsync(T arg, object userState)
    {
      RunAsync(obj => DoWork(arg, userState));
    }

    void DoWork(T arg, object userState)
    {
      Exception error = null;
      try
      {
        _action(arg);
      }
      catch (Exception ex)
      {
        error = ex;
      }
      SynchronizationContext.Send(obj => RaiseServiceCompleted(null, new AsyncCompletedEventArgs(error, false, userState)), null);
    }

    readonly Action<T> _action;
  }

  /// <summary>
  /// Encapsulates a service method that takes two parameters and does not return a value.
  /// </summary>
  public class AsyncServiceAction<T1, T2> : AsyncServiceActionBase
  {
    public AsyncServiceAction(Action<T1, T2> action)
    {
      _action = action;
    }

    public void InvokeServiceAsync(T1 arg1, T2 arg2)
    {
      InvokeServiceAsync(arg1, arg2, null);
    }

    public void InvokeServiceAsync(T1 arg1, T2 arg2, object userState)
    {
      RunAsync(obj => DoWork(arg1, arg2, userState));
    }

    void DoWork(T1 arg1, T2 arg2, object userState)
    {
      Exception error = null;
      try
      {
        _action(arg1, arg2);
      }
      catch (Exception ex)
      {
        error = ex;
      }
      SynchronizationContext.Send(obj => RaiseServiceCompleted(null, new AsyncCompletedEventArgs(error, false, userState)), null);
    }

    readonly Action<T1, T2> _action;
  }

  /// <summary>
  /// Encapsulates a service method that takes three parameters and does not return a value.
  /// </summary>
  public class AsyncServiceAction<T1, T2, T3> : AsyncServiceActionBase
  {
    public AsyncServiceAction(Action<T1, T2, T3> action)
    {
      _action = action;
    }

    public void InvokeServiceAsync(T1 arg1, T2 arg2, T3 arg3)
    {
      InvokeServiceAsync(arg1, arg2, arg3, null);
    }

    public void InvokeServiceAsync(T1 arg1, T2 arg2, T3 arg3, object userState)
    {
      RunAsync(obj => DoWork(arg1, arg2, arg3, userState));
    }

    void DoWork(T1 arg1, T2 arg2, T3 arg3, object userState)
    {
      Exception error = null;
      try
      {
        _action(arg1, arg2, arg3);
      }
      catch (Exception ex)
      {
        error = ex;
      }
      SynchronizationContext.Send(obj => RaiseServiceCompleted(null, new AsyncCompletedEventArgs(error, false, userState)), null);
    }

    readonly Action<T1, T2, T3> _action;
  }

  /// <summary>
  /// Encapsulates a service method that takes four parameters and does not return a value.
  /// </summary>
  public class AsyncServiceAction<T1, T2, T3, T4> : AsyncServiceActionBase
  {
    public AsyncServiceAction(Action<T1, T2, T3, T4> action)
    {
      _action = action;
    }

    public void InvokeServiceAsync(T1 arg1, T2 arg2, T3 arg3, T4 arg4)
    {
      InvokeServiceAsync(arg1, arg2, arg3, arg4, null);
    }

    public void InvokeServiceAsync(T1 arg1, T2 arg2, T3 arg3, T4 arg4, object userState)
    {
      RunAsync(obj => DoWork(arg1, arg2, arg3, arg4, userState));
    }

    void DoWork(T1 arg1, T2 arg2, T3 arg3, T4 arg4, object userState)
    {
      Exception error = null;
      try
      {
        _action(arg1, arg2, arg3, arg4);
      }
      catch (Exception ex)
      {
        error = ex;
      }
      SynchronizationContext.Send(obj => RaiseServiceCompleted(null, new AsyncCompletedEventArgs(error, false, userState)), null);
    }

    readonly Action<T1, T2, T3, T4> _action;
  }

  /// <summary>
  /// Provides data for the ServiceCompleted event.
  /// </summary>
  public class AsyncServiceCompletedEventArgs<TResult> : AsyncCompletedEventArgs
  {
    public AsyncServiceCompletedEventArgs(TResult result, Exception error, bool cancelled, object userState)
      : base(error, cancelled, userState)
    {
      _result = result;
    }

    /// <summary>
    /// Gets the result if the service succeeded successfully. Otherwise an exception in thrown.
    /// </summary>
    public TResult Result
    {
      get
      {
        RaiseExceptionIfNecessary();
        return _result;
      }
    }

    /// <summary>
    /// Gets a value indicating whether the service call is succeeded.
    /// </summary>
    public bool Succeeded
    {
      get { return Error == null && !Cancelled; }
    }

    readonly TResult _result;
  }
}
