﻿using System;
using System.Collections.Generic;
using System.Diagnostics;
using System.Threading;
using System.ComponentModel;
#if SILVERLIGHT
using System.Windows;
#endif

namespace Frmwrk.ComponentModel
{
  /// <summary>
  /// Helper class for synchronized asynchron WCF service calls.
  /// </summary>
  public class SyncCallHelper<T> where T : AsyncCompletedEventArgs  // rename to: SynchronizedServiceCall
  {
    /// <summary>
    /// Initializes a new instance of the <see cref="SyncCallHelper&lt;T&gt;"/> class
    /// and sets the event semaphore to non-signaled.
    /// </summary>
    public SyncCallHelper()
    {
      CheckBackgroundThread();
      Event = new ManualResetEvent(false);
    }

    /// <summary>
    /// Blocks the current thread until the current System.Threading.WaitHandle receives a signal.
    /// </summary>
    public bool Wait()
    {
      CheckBackgroundThread();
      return Event.WaitOne();
    }

    /// <summary>
    /// Blocks the current thread until the current System.Threading.WaitHandle receives a signal, using 32-bit signed integer to measure the time interval.
    /// </summary>
    public bool Wait(int millisecondsTimeout)
    {
      CheckBackgroundThread();
      return Event.WaitOne(millisecondsTimeout);
    }

    /// <summary>
    /// Blocks the current thread until the current instance receives a signal, using a System.TimeSpan to measure the time interval.
    /// </summary>
    public bool Wait(TimeSpan timeout)
    {
      CheckBackgroundThread();
      return Event.WaitOne(timeout);
    }

    /// <summary>
    /// Gets a value indicating whether the service call is succeeded.
    /// </summary>
    public bool Succeeded
    {
      get { return EventArgs != null && EventArgs.Error == null && !EventArgs.Cancelled; }
    }

    /// <summary>
    /// Blocks the current thread until the aynchronous operation completed successful, or throws an exception.
    /// </summary>
    public void WaitUntilSuccesOrFail()
    {
      WaitUntilSuccesOrFail(-1);
    }

    /// <summary>
    /// Blocks the current thread until the aynchronous operation completed successful, or throws an exception.
    /// </summary>
    public void WaitUntilSuccesOrFail(int milliseconds)
    {
      CheckBackgroundThread();

      if (Event.WaitOne(milliseconds))
      {
        if (EventArgs.Error != null)
          throw EventArgs.Error;

        if (EventArgs.Cancelled)
          throw new InvalidOperationException("Service call cancelled.");
      }
      else
        throw new TimeoutException("Service call timed out.");
    }

    /// <summary>
    /// Blocks the current thread until the aynchronous operation completed successful, or throws an exception.
    /// </summary>
    public void WaitUntilSuccesOrFail(TimeSpan timeSpan)
    {
      WaitUntilSuccesOrFail(timeSpan.Milliseconds);
    }

    /// <summary>
    /// Gets the event semaphore used to wait for completion of asynchronous WCF service calls.
    /// </summary>
    public ManualResetEvent Event { get; private set; }

    /// <summary>
    /// Gets the event args that contains the result of the WCF service call.
    /// Is null until the WCF service call has completed.
    /// </summary>
    public T EventArgs { get; private set; }

    /// <summary>
    /// Called when the WCF service call has completed. It is executed by UI-thread anyway.
    /// </summary>
    public void OnCompleted(object sender, T e)
    {
      var res = (SyncCallHelper<T>)e.UserState;
      Debug.Assert(res!=null,"Pass instance of this class to '...Async' function.");
      res.EventArgs = e;
      res.Event.Set();
    }

    /// <summary>
    /// Checks that current thread is the background thread.
    /// Throws InvalidOperationException if the current thread is the UI thread.
    /// </summary>
    static void CheckBackgroundThread()
    {
#if SILVERLIGHT
      if (Deployment.Current.CheckAccess())
        throw new InvalidOperationException("SyncCallHelper must only be used in background thread.");
#endif
    }
  }
}
